--Instatiates a Block RAM. 
-- Calculates write and read addresses. The read address is always one ahead the write address. 
-- The RAM is n elements long to store n-1 elements

 
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use ieee.math_real.all;

ENTITY fifowrapper IS
	GENERIC (
		DATAWIDTH : integer := 8;	--Upon component instantiation, these values are overwritten
		ADDRLEN : integer := 127
	);
	PORT (
		clk : in std_logic;
		reset : in std_logic;
		wdata : in std_logic_vector(DATAWIDTH - 1 DOWNTO 0);
		--rdatatop : out std_logic_vector(DATAWIDTH - 1 DOWNTO 0)
		rdatatop : out unsigned(DATAWIDTH - 1 DOWNTO 0)
	);
	
END fifowrapper;

ARCHITECTURE Behavioral OF fifowrapper IS

	COMPONENT fifo 
		GENERIC(
			DATA : integer := DATAWIDTH;		--use this as data bit width
			ADDRLEN : integer := ADDRLEN
		);
		PORT(
			clk : in std_logic;
			we : in std_logic;		--write enable
			waddr : in std_logic_vector(integer(ceil(log2(real(ADDRLEN))))-1 downto 0);	--for correct FIFO operation, the read pointer must always be ahead the write pointer.
			wdata : in std_logic_vector((DATAWIDTH - 1) DOWNTO 0);		
			raddr : in std_logic_vector(integer(ceil(log2(real(ADDRLEN))))-1 downto 0);	--therefore, the FIFO must be one data cell longer than it must actually hold
			--rdata : out std_logic_vector((DATAWIDTH - 1) DOWNTO 0)
			rdata : out unsigned((DATAWIDTH - 1) DOWNTO 0)
		);
	END COMPONENT;
		
	signal writeaddress, writeaddress_next : unsigned(integer(ceil(log2(real(ADDRLEN))))-1 downto 0);	--If the FIFO must store 256 words, actually 257 cells are required
	signal readaddress, readaddress_next : unsigned(integer(ceil(log2(real(ADDRLEN)))) - 1 DOWNTO 0);
	signal writeenable : std_logic;
		
BEGIN
	myfifo : fifo
		PORT MAP(
			clk => clk,
			we => writeenable,
			waddr => std_logic_vector(writeaddress),
			wdata => wdata,
			raddr => std_logic_vector(readaddress),
			rdata => rdatatop
		);
	
	clockproc : PROCESS(clk, reset)
	BEGIN
		IF reset = '1' THEN
			writeenable <= '0';
			writeaddress <= (OTHERS => '0');
			readaddress(0) <= '1';				--Ensure that read address is alway one cell before write address! Because of clock synchronous design, this is always valid, if both addresses are incremented by one
			readaddress(integer(ceil(log2(real(ADDRLEN)))) - 1 DOWNTO 1) <= (OTHERS => '0');
		ELSIF clk'event AND clk = '1' THEN
			writeaddress <= writeaddress_next;
			readaddress <= readaddress_next;
			writeenable <= '1';
		END IF;
	END PROCESS;
	
	writeproc : PROCESS(writeaddress, readaddress)
	BEGIN
		
		--Wrap write and read addresses at end of FIFO
		IF writeaddress = ADDRLEN THEN
			writeaddress_next <= (OTHERS => '0');
		ELSE
			writeaddress_next <= writeaddress + 1;
		END IF;
		
		IF readaddress = ADDRLEN THEN
			readaddress_next <= (OTHERS => '0');
		ELSE
			readaddress_next <= readaddress + 1;
		END IF;
	
	END PROCESS;

END Behavioral;

